# Autogenerated by devscripts/update-version.py

__version__ = '2024.04.09'

RELEASE_GIT_HEAD = 'ff07792676f404ffff6ee61b5638c9dc1a33a37a'

VARIANT = None

UPDATE_HINT = None

CHANNEL = 'stable'

ORIGIN = 'yt-dlp/yt-dlp'

_pkg_version = '2024.04.09'
